/*
 * $Id: share_hash.c,v 1.10 2003/03/06 02:31:59 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gift.h"

#include "share_file.h"
#include "share_hash.h"

#include "file.h"

/*****************************************************************************/

static Dataset *algorithms = NULL;

/*****************************************************************************/

ShareHash *share_hash_new (unsigned char *hash, int len, char *type)
{
	ShareHash *sh;

	if (!hash || len <= 0)
		return NULL;

	if (!(sh = malloc (sizeof (ShareHash))))
		return NULL;

	sh->hash = hash;
	sh->len  = len;
	sh->type = type;

	return sh;
}

ShareHash *share_hash_dup (ShareHash *sh)
{
	ShareHash *dup;

	if (!sh || !(dup = share_hash_new (sh->hash, sh->len, sh->type)))
		return NULL;

	dup->hash = malloc (sh->len);
	memcpy (dup->hash, sh->hash, sh->len);

	return dup;
}

void share_hash_free (ShareHash *sh)
{
	if (!sh)
		return;

	free (sh->hash);
	free (sh);
}

/*****************************************************************************/

int hash_algo_register (Protocol *p, char *type,
                        HashAlgorithm algo, HashHuman human)
{
	ShareHashType *hash_type;

	if (!p || !type || !algo || !human)
		return FALSE;

	if ((hash_type = dataset_lookup (algorithms, type, STRLEN_0 (type))))
	{
		dataset_insert (&hash_type->ref, &p, sizeof (p), "hash_algo_t", 0);
		return FALSE;
	}

	if (!(hash_type = malloc (sizeof (ShareHashType))))
		return FALSE;

	memset (hash_type, 0, sizeof (ShareHashType));

	hash_type->algo    = algo;
	hash_type->human   = human;
	hash_type->type    = STRDUP (type);

	dataset_insert (&hash_type->ref, &p, sizeof (p), "hash_algo_t", 0);
	dataset_insert (&algorithms, type, STRLEN_0 (type), hash_type, 0);

	return TRUE;
}

void hash_algo_unregister (Protocol *p, char *type)
{
	ShareHashType *hash_type;

	if (!(hash_type = hash_algo_lookup (type)))
		return;

	dataset_remove (hash_type->ref, &p, sizeof (p));

	/* don't unregister if there are still protocols with references */
	if (dataset_length (hash_type->ref) > 0)
		return;

	dataset_clear (hash_type->ref);

	dataset_remove (algorithms, type, STRLEN_0 (type));

	free (hash_type->type);
	free (hash_type);
}

ShareHashType *hash_algo_lookup (char *type)
{
	return dataset_lookup (algorithms, type, STRLEN_0 (type));
}

/*****************************************************************************/

static int algo_calc (Dataset *d, DatasetNode *node, FileShare *file)
{
	ShareHashType *hash_type = node->value;
	char          *type      = hash_type->type;
	unsigned char *hash      = NULL;
	int            len       = 0;
	char          *host_path;

	/* check if we already have a hash stored here */
	if (share_hash_get (file, type))
		return FALSE;

	if ((host_path = file_host_path (SHARE_DATA(file)->path)))
	{
		hash = hash_type->algo (host_path, type, &len);
		free (host_path);
	}

	share_hash_set (file, type, hash, len);
	return FALSE;
}

int hash_algo_run (FileShare *file)
{
	if (!file || file->p)
	{
		TRACE (("eep!  what are you doing!?"));
		return FALSE;
	}

	dataset_foreach (algorithms, DATASET_FOREACH (algo_calc), file);
	return TRUE;
}

/*****************************************************************************/

static int get_last (Dataset *d, DatasetNode *node, void *udata)
{
	if (!hash_algo_lookup (node->key))
		return FALSE;

	return TRUE;
}

ShareHash *share_hash_get (FileShare *file, char *type)
{
	ShareHash *sh;

	if (!file)
		return NULL;

	if (file->p)
		sh = SHARE_DATA(file)->hash.hash;
	else
	{
		if (!type)
		{
			/* Select the first registered hashing algorithm.
			 * This actually selects the last registered due to the way
			 * the dataset is currently stored. */
			sh = dataset_find (SHARE_DATA(file)->hash.hashes, 
			                   DATASET_FOREACH (get_last), NULL);
		}
		else
		{
			sh = dataset_lookup (SHARE_DATA(file)->hash.hashes,
			                     type, STRLEN_0 (type));
		}
	}

	return sh;
}

static int get_key (Dataset *d, DatasetNode *node, void **args)
{
	char   *type   = args[0];
	char  **r_type = args[1];

	if (strcmp (node->key, type) != 0)
		return FALSE;

	*r_type = node->key;
	return TRUE;
}

/* Get the value of the hash_type key in the dataset. */
static char *find_key (Dataset *hashes, char *type)
{
	char *hash_type;
	void *args[] = { type, &hash_type };

	dataset_find (hashes, DATASET_FOREACH (get_key), args);

	return hash_type;
}

int share_hash_set (FileShare *file, char *type,
                    unsigned char *hash, int len)
{
	ShareHash *sh;

	if (!file || !type)
		return FALSE;

	assert (share_hash_get (file, type) == NULL);

	if (!(sh = share_hash_new (hash, len, type)))
		return FALSE;

	if (file->p)
		SHARE_DATA(file)->hash.hash = sh;
	else
	{
		dataset_insert (&SHARE_DATA(file)->hash.hashes,
		                type, STRLEN_0 (type), sh, 0);
		
		/* 
		 * The type string may be dynamically allocated. Since we don't
		 * want to have to free sh->type in share_hash_free, share it with
		 * the dataset. NOTE: protocols should always pass in a literal
		 * constant string ("MD5") for type, and so don't need this.
		 */
		sh->type = find_key (SHARE_DATA(file)->hash.hashes, type);
	}

	return TRUE;
}

char *share_hash_human (FileShare *file, char *type, int *r_len)
{
	ShareHashType *hash_type;
	ShareHash     *sh;
	char          *hash;
	int            len;

	if (!(sh = share_hash_get (file, type)))
		return NULL;

	if (!type)
		type = sh->type;

	if (!(hash_type = dataset_lookup (algorithms, type, STRLEN_0 (type))))
		return NULL;

	hash = hash_type->human (sh->hash, &len);

	if (r_len)
		*r_len = len;

	return hash;
}

static int sh_clear (Dataset *d, DatasetNode *node, FileShare *file)
{
	ShareHash *sh = node->value;

	share_hash_free (sh);
	return TRUE;
}

void share_hash_clear (FileShare *file)
{
	if (!file || !SHARE_DATA(file))
		return;

	if (file->p)
		share_hash_free (SHARE_DATA(file)->hash.hash);
	else
	{
		dataset_foreach (SHARE_DATA(file)->hash.hashes,
						 DATASET_FOREACH (sh_clear), file);
		dataset_clear (SHARE_DATA(file)->hash.hashes);
	}

	memset (&SHARE_DATA(file)->hash, 0, sizeof (SHARE_DATA(file)->hash));
}

void share_hash_foreach (FileShare *file, DatasetForeach func, void *udata)
{
	if (!file || !func)
		return;

	assert (file->p == NULL);          /* TODO */

	dataset_foreach (SHARE_DATA(file)->hash.hashes, func, udata);
}

/*****************************************************************************/

/* Return the algorithm used to create a hash string. */
char *hash_algo (char *hash)
{
	char       *colon;
	size_t      len;
	static char algo[32];

	if (!hash || !(colon = strchr (hash, ':')))
		return NULL;

	len = CLAMP (colon - hash, 0, sizeof (algo) - 1);

	memcpy (algo, hash, len);
	algo[len] = 0;

	return algo;
}

/* Skip the algorithm prefix ("MD5:") and return the hash data. */
char *hash_data (char *hash)
{
	char *data;

	if (!hash || !(data = strchr (hash, ':')))
		return hash;

	return data + 1;
}
